@@ -3,6 +3,7 @@
 #include "nix/util/signals.hh"
 #include "nix/store/posix-fs-canonicalise.hh"
 #include "nix/util/posix-source-accessor.hh"
+#include "nix/util/file-system.hh"
 
 #include <cstdlib>
 #include <cstring>
@@ -17,19 +18,18 @@
 
 namespace nix {
 
-static void makeWritable(const Path & path)
+static void makeWritable(const std::filesystem::path & path)
 {
     auto st = lstat(path);
-    if (chmod(path.c_str(), st.st_mode | S_IWUSR) == -1)
-        throw SysError("changing writability of '%1%'", path);
+    chmod(path, st.st_mode | S_IWUSR);
 }
 
 struct MakeReadOnly
 {
-    Path path;
+    std::filesystem::path path;
 
-    MakeReadOnly(const PathView path)
-        : path(path)
+    MakeReadOnly(std::filesystem::path path)
+        : path(std::move(path))
     {
     }
 
@@ -37,8 +37,8 @@ struct MakeReadOnly
     {
         try {
             /* This will make the path read-only. */
-            if (path != "")
-                canonicaliseTimestampAndPermissions(path);
+            if (!path.empty())
+                canonicaliseTimestampAndPermissions(path.string());
         } catch (...) {
             ignoreExceptionInDestructor();
         }
@@ -50,9 +50,9 @@ LocalStore::InodeHash LocalStore::loadInodeHash()
     debug("loading hash inodes in memory");
     InodeHash inodeHash;
 
-    AutoCloseDir dir(opendir(linksDir.c_str()));
+    AutoCloseDir dir(opendir(linksDir.string().c_str()));
     if (!dir)
-        throw SysError("opening directory '%1%'", linksDir);
+        throw SysError("opening directory %1%", PathFmt(linksDir));
 
     struct dirent * dirent;
     while (errno = 0, dirent = readdir(dir.get())) { /* sic */
@@ -61,20 +61,20 @@ LocalStore::InodeHash LocalStore::loadInodeHash()
         inodeHash.insert(dirent->d_ino);
     }
     if (errno)
-        throw SysError("reading directory '%1%'", linksDir);
+        throw SysError("reading directory %1%", PathFmt(linksDir));
 
     printMsg(lvlTalkative, "loaded %1% hash inodes", inodeHash.size());
 
     return inodeHash;
 }
 
-Strings LocalStore::readDirectoryIgnoringInodes(const Path & path, const InodeHash & inodeHash)
+Strings LocalStore::readDirectoryIgnoringInodes(const std::filesystem::path & path, const InodeHash & inodeHash)
 {
     Strings names;
 
-    AutoCloseDir dir(opendir(path.c_str()));
+    AutoCloseDir dir(opendir(path.string().c_str()));
     if (!dir)
-        throw SysError("opening directory '%1%'", path);
+        throw SysError("opening directory %s", PathFmt(path));
 
     struct dirent * dirent;
     while (errno = 0, dirent = readdir(dir.get())) { /* sic */
@@ -91,13 +91,13 @@ Strings LocalStore::readDirectoryIgnoringInodes(const Path & path, const InodeHa
         names.push_back(name);
     }
     if (errno)
-        throw SysError("reading directory '%1%'", path);
+        throw SysError("reading directory %s", PathFmt(path));
 
     return names;
 }
 
 void LocalStore::optimisePath_(
-    Activity * act, OptimiseStats & stats, const Path & path, InodeHash & inodeHash, RepairFlag repair)
+    Activity * act, OptimiseStats & stats, const std::filesystem::path & path, InodeHash & inodeHash, RepairFlag repair)
 {
     checkInterrupt();
 
@@ -110,8 +110,8 @@ void LocalStore::optimisePath_(
        See https://github.com/NixOS/nix/issues/1443 and
        https://github.com/NixOS/nix/pull/2230 for more discussion. */
 
-    if (std::regex_search(path, std::regex("\\.app/Contents/.+$"))) {
-        debug("'%1%' is not allowed to be linked in macOS", path);
+    if (std::regex_search(path.string(), std::regex("\\.app/Contents/.+$"))) {
+        debug("%s is not allowed to be linked in macOS", PathFmt(path));
         return;
     }
 #endif
@@ -119,7 +119,7 @@ void LocalStore::optimisePath_(
     if (S_ISDIR(st.st_mode)) {
         Strings names = readDirectoryIgnoringInodes(path, inodeHash);
         for (auto & i : names)
-            optimisePath_(act, stats, path + "/" + i, inodeHash, repair);
+            optimisePath_(act, stats, path / i, inodeHash, repair);
         return;
     }
 
@@ -136,13 +136,13 @@ void LocalStore::optimisePath_(
        NixOS (example: $fontconfig/var/cache being modified).  Skip
        those files.  FIXME: check the modification time. */
     if (S_ISREG(st.st_mode) && (st.st_mode & S_IWUSR)) {
-        warn("skipping suspicious writable file '%1%'", path);
+        warn("skipping suspicious writable file '%s'", PathFmt(path));
         return;
     }
 
     /* This can still happen on top-level files. */
     if (st.st_nlink > 1 && inodeHash.count(st.st_ino)) {
-        debug("'%s' is already linked, with %d other file(s)", path, st.st_nlink - 2);
+        debug("%s is already linked, with %d other file(s)", PathFmt(path), st.st_nlink - 2);
         return;
     }
 
@@ -157,19 +157,19 @@ void LocalStore::optimisePath_(
        the contents of the target (which may not even exist). */
     Hash hash = ({
         hashPath(
-            {make_ref<PosixSourceAccessor>(), CanonPath(path)},
+            {make_ref<PosixSourceAccessor>(), CanonPath(path.string())},
             FileSerialisationMethod::NixArchive,
             HashAlgorithm::SHA256)
             .hash;
     });
-    debug("'%1%' has hash '%2%'", path, hash.to_string(HashFormat::Nix32, true));
+    debug("%s has hash '%s'", PathFmt(path), hash.to_string(HashFormat::Nix32, true));
 
     /* Check if this is a known hash. */
     std::filesystem::path linkPath = std::filesystem::path{linksDir} / hash.to_string(HashFormat::Nix32, false);
 
     /* Maybe delete the link, if it has been corrupted. */
     if (std::filesystem::exists(std::filesystem::symlink_status(linkPath))) {
-        auto stLink = lstat(linkPath.string());
+        auto stLink = lstat(linkPath);
         if (st.st_size != stLink.st_size || (repair && hash != ({
                                                            hashPath(
                                                                makeFSSourceAccessor(linkPath),
@@ -178,7 +178,7 @@ void LocalStore::optimisePath_(
                                                                .hash;
                                                        }))) {
             // XXX: Consider overwriting linkPath with our valid version.
-            warn("removing corrupted link %s", linkPath);
+            warn("removing corrupted link %s", PathFmt(linkPath));
             warn(
                 "There may be more corrupted paths."
                 "\nYou should run `nix-store --verify --check-contents --repair` to fix them all");
@@ -201,38 +201,39 @@ void LocalStore::optimisePath_(
                 /* On ext4, that probably means the directory index is
                    full.  When that happens, it's fine to ignore it: we
                    just effectively disable deduplication of this
-                   file.  */
-                printInfo("cannot link %s to '%s': %s", linkPath, path, strerror(errno));
+                   file.
+                   */
+                printInfo("cannot link %s to '%s': %s", PathFmt(linkPath), PathFmt(path), e.code().message());
                 return;
             }
 
             else
-                throw;
+                throw SystemError(e.code(), "creating hard link from %1% to %2%", PathFmt(linkPath), PathFmt(path));
         }
     }
 
     /* Yes!  We've seen a file with the same contents.  Replace the
        current file with a hard link to that file. */
-    auto stLink = lstat(linkPath.string());
+    auto stLink = lstat(linkPath);
 
     if (st.st_ino == stLink.st_ino) {
-        debug("'%1%' is already linked to %2%", path, linkPath);
+        debug("%1% is already linked to %2%", PathFmt(path), PathFmt(linkPath));
         return;
     }
 
-    printMsg(lvlTalkative, "linking '%1%' to %2%", path, linkPath);
+    printMsg(lvlTalkative, "linking %1% to %2%", PathFmt(path), PathFmt(linkPath));
 
     /* Make the containing directory writable, but only if it's not
        the store itself (we don't want or need to mess with its
        permissions). */
-    const Path dirOfPath(dirOf(path));
+    const auto dirOfPath = path.parent_path();
     bool mustToggle = dirOfPath != config->realStoreDir.get();
     if (mustToggle)
         makeWritable(dirOfPath);
 
     /* When we're done, make the directory read-only again and reset
        its timestamp back to 0. */
-    MakeReadOnly makeReadOnly(mustToggle ? dirOfPath : "");
+    MakeReadOnly makeReadOnly(mustToggle ? dirOfPath : std::filesystem::path{});
 
     std::filesystem::path tempLink = makeTempPath(config->realStoreDir.get(), ".tmp-link");
 
@@ -245,10 +246,10 @@ void LocalStore::optimisePath_(
                systems).  This is likely to happen with empty files.
                Just shrug and ignore. */
             if (st.st_size)
-                printInfo("%1% has maximum number of links", linkPath);
+                printInfo("%1% has maximum number of links", PathFmt(linkPath));
             return;
         }
-        throw;
+        throw SystemError(e.code(), "creating hard link from %1% to %2%", PathFmt(linkPath), PathFmt(tempLink));
     }
 
     /* Atomically replace the old file with the new hard link. */
@@ -259,17 +260,17 @@ void LocalStore::optimisePath_(
             std::error_code ec;
             remove(tempLink, ec); /* Clean up after ourselves. */
             if (ec)
-                printError("unable to unlink %1%: %2%", tempLink, ec.message());
+                printError("unable to unlink %1%: %2%", PathFmt(tempLink), ec.message());
         }
         if (e.code() == std::errc::too_many_links) {
             /* Some filesystems generate too many links on the rename,
                rather than on the original link.  (Probably it
                temporarily increases the st_nlink field before
                decreasing it again.) */
-            debug("%s has reached maximum number of links", linkPath);
+            debug("%s has reached maximum number of links", PathFmt(linkPath));
             return;
         }
-        throw;
+        throw SystemError(e.code(), "renaming %1% to %2%", PathFmt(tempLink), PathFmt(path));
     }
 
     stats.filesLinked++;
@@ -303,7 +304,7 @@ void LocalStore::optimiseStore(OptimiseStats & stats)
             continue; /* path was GC'ed, probably */
         {
             Activity act(*logger, lvlTalkative, actUnknown, fmt("optimising path '%s'", printStorePath(i)));
-            optimisePath_(&act, stats, config->realStoreDir + "/" + std::string(i.to_string()), inodeHash, NoRepair);
+            optimisePath_(&act, stats, config->realStoreDir.get() / i.to_string(), inodeHash, NoRepair);
         }
         done++;
         act.progress(done, paths.size());
@@ -319,12 +320,12 @@ void LocalStore::optimiseStore()
     printInfo("%s freed by hard-linking %d files", renderSize(stats.bytesFreed), stats.filesLinked);
 }
 
-void LocalStore::optimisePath(const Path & path, RepairFlag repair)
+void LocalStore::optimisePath(const std::filesystem::path & path, RepairFlag repair)
 {
     OptimiseStats stats;
     InodeHash inodeHash;
 
-    if (settings.autoOptimiseStore)
+    if (config->getLocalSettings().autoOptimiseStore)
         optimisePath_(nullptr, stats, path, inodeHash, repair);
 }
 
