@@ -3,6 +3,7 @@
 #include "nix/store/store-api.hh"
 #include "nix/store/gc-store.hh"
 #include "nix/store/build-result.hh"
+#include "nix/store/common-protocol.hh"
 #include "nix/store/worker-protocol.hh"
 #include "nix/store/worker-protocol-impl.hh"
 #include "nix/util/archive.hh"
@@ -13,6 +14,37 @@
 
 namespace nix {
 
+const WorkerProto::Version WorkerProto::latest = {
+    .number =
+        {
+            .major = 1,
+            .minor = 38,
+        },
+};
+
+const WorkerProto::Version WorkerProto::minimum = {
+    .number =
+        {
+            .major = 1,
+            .minor = 18,
+        },
+};
+
+std::partial_ordering WorkerProto::Version::operator<=>(const WorkerProto::Version & other) const
+{
+    auto numCmp = number <=> other.number;
+    bool thisSubsetEq = std::includes(other.features.begin(), other.features.end(), features.begin(), features.end());
+    bool otherSubsetEq = std::includes(features.begin(), features.end(), other.features.begin(), other.features.end());
+
+    if (numCmp == 0 && thisSubsetEq && otherSubsetEq)
+        return std::partial_ordering::equivalent;
+    if (numCmp <= 0 && thisSubsetEq)
+        return std::partial_ordering::less;
+    if (numCmp >= 0 && otherSubsetEq)
+        return std::partial_ordering::greater;
+    return std::partial_ordering::unordered;
+}
+
 /* protocol-specific definitions */
 
 BuildMode WorkerProto::Serialise<BuildMode>::read(const StoreDirConfig & store, WorkerProto::ReadConn conn)
@@ -152,7 +184,7 @@ void WorkerProto::Serialise<std::optional<std::chrono::microseconds>>::write(
 DerivedPath WorkerProto::Serialise<DerivedPath>::read(const StoreDirConfig & store, WorkerProto::ReadConn conn)
 {
     auto s = readString(conn.from);
-    if (GET_PROTOCOL_MINOR(conn.version) >= 30) {
+    if (conn.version >= WorkerProto::Version{.number = {1, 30}}) {
         return DerivedPath::parseLegacy(store, s);
     } else {
         return parsePathWithOutputs(store, s).toDerivedPath();
@@ -162,7 +194,7 @@ DerivedPath WorkerProto::Serialise<DerivedPath>::read(const StoreDirConfig & sto
 void WorkerProto::Serialise<DerivedPath>::write(
     const StoreDirConfig & store, WorkerProto::WriteConn conn, const DerivedPath & req)
 {
-    if (GET_PROTOCOL_MINOR(conn.version) >= 30) {
+    if (conn.version >= WorkerProto::Version{.number = {1, 30}}) {
         conn.to << req.to_string_legacy(store);
     } else {
         auto sOrDrvPath = StorePathWithOutputs::tryFromDerivedPath(req);
@@ -173,8 +205,8 @@ void WorkerProto::Serialise<DerivedPath>::write(
                     throw Error(
                         "trying to request '%s', but daemon protocol %d.%d is too old (< 1.29) to request a derivation file",
                         store.printStorePath(drvPath),
-                        GET_PROTOCOL_MAJOR(conn.version),
-                        GET_PROTOCOL_MINOR(conn.version));
+                        conn.version.number.major,
+                        conn.version.number.minor);
                 },
                 [&](std::monostate) {
                     throw Error(
@@ -207,31 +239,42 @@ BuildResult WorkerProto::Serialise<BuildResult>::read(const StoreDirConfig & sto
 {
     BuildResult res;
     BuildResult::Success success;
-    BuildResult::Failure failure;
 
-    auto rawStatus = readInt(conn.from);
-    conn.from >> failure.errorMsg;
+    // Temp variables for failure fields since BuildError uses methods
+    std::string errorMsg;
+    bool isNonDeterministic = false;
 
-    if (GET_PROTOCOL_MINOR(conn.version) >= 29) {
-        conn.from >> res.timesBuilt >> failure.isNonDeterministic >> res.startTime >> res.stopTime;
+    auto status = WorkerProto::Serialise<BuildResultStatus>::read(store, conn);
+    conn.from >> errorMsg;
+
+    if (conn.version >= WorkerProto::Version{.number = {1, 29}}) {
+        conn.from >> res.timesBuilt >> isNonDeterministic >> res.startTime >> res.stopTime;
     }
-    if (GET_PROTOCOL_MINOR(conn.version) >= 37) {
+    if (conn.version >= WorkerProto::Version{.number = {1, 37}}) {
         res.cpuUser = WorkerProto::Serialise<std::optional<std::chrono::microseconds>>::read(store, conn);
         res.cpuSystem = WorkerProto::Serialise<std::optional<std::chrono::microseconds>>::read(store, conn);
     }
-    if (GET_PROTOCOL_MINOR(conn.version) >= 28) {
+    if (conn.version >= WorkerProto::Version{.number = {1, 28}}) {
         auto builtOutputs = WorkerProto::Serialise<DrvOutputs>::read(store, conn);
         for (auto && [output, realisation] : builtOutputs)
             success.builtOutputs.insert_or_assign(std::move(output.outputName), std::move(realisation));
     }
 
-    if (BuildResult::Success::statusIs(rawStatus)) {
-        success.status = static_cast<BuildResult::Success::Status>(rawStatus);
-        res.inner = std::move(success);
-    } else {
-        failure.status = static_cast<BuildResult::Failure::Status>(rawStatus);
-        res.inner = std::move(failure);
-    }
+    res.inner = std::visit(
+        overloaded{
+            [&](BuildResult::Success::Status s) -> decltype(res.inner) {
+                success.status = s;
+                return std::move(success);
+            },
+            [&](BuildResult::Failure::Status s) -> decltype(res.inner) {
+                return BuildResult::Failure{{
+                    .status = s,
+                    .msg = HintFmt(std::move(errorMsg)),
+                    .isNonDeterministic = isNonDeterministic,
+                }};
+            },
+        },
+        status);
 
     return res;
 }
@@ -246,14 +289,14 @@ void WorkerProto::Serialise<BuildResult>::write(
        default value for the fields that don't exist in that case. */
     auto common = [&](std::string_view errorMsg, bool isNonDeterministic, const auto & builtOutputs) {
         conn.to << errorMsg;
-        if (GET_PROTOCOL_MINOR(conn.version) >= 29) {
+        if (conn.version >= WorkerProto::Version{.number = {1, 29}}) {
             conn.to << res.timesBuilt << isNonDeterministic << res.startTime << res.stopTime;
         }
-        if (GET_PROTOCOL_MINOR(conn.version) >= 37) {
+        if (conn.version >= WorkerProto::Version{.number = {1, 37}}) {
             WorkerProto::write(store, conn, res.cpuUser);
             WorkerProto::write(store, conn, res.cpuSystem);
         }
-        if (GET_PROTOCOL_MINOR(conn.version) >= 28) {
+        if (conn.version >= WorkerProto::Version{.number = {1, 28}}) {
             DrvOutputs builtOutputsFullKey;
             for (auto & [output, realisation] : builtOutputs)
                 builtOutputsFullKey.insert_or_assign(realisation.id, realisation);
@@ -263,11 +306,11 @@ void WorkerProto::Serialise<BuildResult>::write(
     std::visit(
         overloaded{
             [&](const BuildResult::Failure & failure) {
-                conn.to << failure.status;
-                common(failure.errorMsg, failure.isNonDeterministic, decltype(BuildResult::Success::builtOutputs){});
+                WorkerProto::write(store, conn, BuildResultStatus{failure.status});
+                common(failure.message(), failure.isNonDeterministic, decltype(BuildResult::Success::builtOutputs){});
             },
             [&](const BuildResult::Success & success) {
-                conn.to << success.status;
+                WorkerProto::write(store, conn, BuildResultStatus{success.status});
                 common(/*errorMsg=*/"", /*isNonDeterministic=*/false, success.builtOutputs);
             },
         },
@@ -298,9 +341,9 @@ UnkeyedValidPathInfo WorkerProto::Serialise<UnkeyedValidPathInfo>::read(const St
     info.deriver = std::move(deriver);
     info.references = WorkerProto::Serialise<StorePathSet>::read(store, conn);
     conn.from >> info.registrationTime >> info.narSize;
-    if (GET_PROTOCOL_MINOR(conn.version) >= 16) {
+    if (conn.version >= WorkerProto::Version{.number = {1, 16}}) {
         conn.from >> info.ultimate;
-        info.sigs = readStrings<StringSet>(conn.from);
+        info.sigs = WorkerProto::Serialise<std::set<Signature>>::read(store, conn);
         info.ca = ContentAddress::parseOpt(readString(conn.from));
     }
     return info;
@@ -313,8 +356,10 @@ void WorkerProto::Serialise<UnkeyedValidPathInfo>::write(
     conn.to << pathInfo.narHash.to_string(HashFormat::Base16, false);
     WorkerProto::write(store, conn, pathInfo.references);
     conn.to << pathInfo.registrationTime << pathInfo.narSize;
-    if (GET_PROTOCOL_MINOR(conn.version) >= 16) {
-        conn.to << pathInfo.ultimate << pathInfo.sigs << renderContentAddress(pathInfo.ca);
+    if (conn.version >= WorkerProto::Version{.number = {1, 16}}) {
+        conn.to << pathInfo.ultimate;
+        WorkerProto::write(store, conn, pathInfo.sigs);
+        conn.to << renderContentAddress(pathInfo.ca);
     }
 }
 
@@ -323,11 +368,11 @@ WorkerProto::Serialise<WorkerProto::ClientHandshakeInfo>::read(const StoreDirCon
 {
     WorkerProto::ClientHandshakeInfo res;
 
-    if (GET_PROTOCOL_MINOR(conn.version) >= 33) {
+    if (conn.version >= WorkerProto::Version{.number = {1, 33}}) {
         res.daemonNixVersion = readString(conn.from);
     }
 
-    if (GET_PROTOCOL_MINOR(conn.version) >= 35) {
+    if (conn.version >= WorkerProto::Version{.number = {1, 35}}) {
         res.remoteTrustsUs = WorkerProto::Serialise<std::optional<TrustedFlag>>::read(store, conn);
     } else {
         // We don't know the answer; protocol to old.
@@ -340,12 +385,12 @@ WorkerProto::Serialise<WorkerProto::ClientHandshakeInfo>::read(const StoreDirCon
 void WorkerProto::Serialise<WorkerProto::ClientHandshakeInfo>::write(
     const StoreDirConfig & store, WriteConn conn, const WorkerProto::ClientHandshakeInfo & info)
 {
-    if (GET_PROTOCOL_MINOR(conn.version) >= 33) {
+    if (conn.version >= WorkerProto::Version{.number = {1, 33}}) {
         assert(info.daemonNixVersion);
         conn.to << *info.daemonNixVersion;
     }
 
-    if (GET_PROTOCOL_MINOR(conn.version) >= 35) {
+    if (conn.version >= WorkerProto::Version{.number = {1, 35}}) {
         WorkerProto::write(store, conn, info.remoteTrustsUs);
     }
 }
