@@ -103,7 +103,7 @@ void Config::getSettings(std::map<std::string, SettingInfo> & res, bool overridd
  */
 static void parseConfigFiles(
     const std::string & contents,
-    const std::string & path,
+    const std::filesystem::path & path,
     std::vector<std::pair<std::string, std::string>> & parsedContents)
 {
     unsigned int pos = 0;
@@ -122,7 +122,7 @@ static void parseConfigFiles(
             continue;
 
         if (tokens.size() < 2)
-            throw UsageError("syntax error in configuration line '%1%' in '%2%'", line, path);
+            throw UsageError("syntax error in configuration line '%s' in %s", line, PathFmt(path));
 
         auto include = false;
         auto ignoreMissing = false;
@@ -135,8 +135,9 @@ static void parseConfigFiles(
 
         if (include) {
             if (tokens.size() != 2)
-                throw UsageError("syntax error in configuration line '%1%' in '%2%'", line, path);
-            auto p = absPath(tokens[1], dirOf(path));
+                throw UsageError("syntax error in configuration line '%1%' in %s", line, PathFmt(path));
+            auto parent = path.parent_path();
+            auto p = absPath(std::filesystem::path{tokens[1]}, &parent);
             if (pathExists(p)) {
                 try {
                     std::string includedContents = readFile(p);
@@ -145,13 +146,13 @@ static void parseConfigFiles(
                     // TODO: Do we actually want to ignore this? Or is it better to fail?
                 }
             } else if (!ignoreMissing) {
-                throw Error("file '%1%' included from '%2%' not found", p, path);
+                throw Error("file %s included from %s not found", PathFmt(p), PathFmt(path));
             }
             continue;
         }
 
         if (tokens[1] != "=")
-            throw UsageError("syntax error in configuration line '%1%' in '%2%'", line, path);
+            throw UsageError("syntax error in configuration line '%s' in %s", line, PathFmt(path));
 
         std::string name = std::move(tokens[0]);
 
@@ -393,6 +394,28 @@ std::string BaseSetting<StringSet>::to_string() const
     return concatStringsSep(" ", value);
 }
 
+template<>
+std::set<std::filesystem::path> BaseSetting<std::set<std::filesystem::path>>::parse(const std::string & str) const
+{
+    auto tokens = tokenizeString<StringSet>(str);
+    return {tokens.begin(), tokens.end()};
+}
+
+template<>
+void BaseSetting<std::set<std::filesystem::path>>::appendOrSet(std::set<std::filesystem::path> newValue, bool append)
+{
+    if (!append)
+        value.clear();
+    value.insert(std::make_move_iterator(newValue.begin()), std::make_move_iterator(newValue.end()));
+}
+
+template<>
+std::string BaseSetting<std::set<std::filesystem::path>>::to_string() const
+{
+    return concatStringsSep(
+        " ", value | std::views::transform([](const auto & p) { return p.string(); }) | std::ranges::to<Strings>());
+}
+
 template<>
 std::set<ExperimentalFeature> BaseSetting<std::set<ExperimentalFeature>>::parse(const std::string & str) const
 {
@@ -402,7 +425,11 @@ std::set<ExperimentalFeature> BaseSetting<std::set<ExperimentalFeature>>::parse(
             res.insert(thisXpFeature.value());
             if (thisXpFeature.value() == Xp::Flakes)
                 res.insert(Xp::FetchTree);
-        } else
+        } else if (s == "no-url-literals")
+            warn(
+                "experimental feature '%s' has been stabilized and renamed; use 'lint-url-literals = fatal' setting instead",
+                s);
+        else
             warn("unknown experimental feature '%s'", s);
     }
     return res;
@@ -456,7 +483,7 @@ std::string BaseSetting<StringMap>::to_string() const
         [](const auto & kvpair) { return kvpair.first + "=" + kvpair.second; });
 }
 
-static Path parsePath(const AbstractSetting & s, const std::string & str)
+static AbsolutePath parseAbsolutePath(const AbstractSetting & s, const std::string & str)
 {
     if (str == "")
         throw UsageError("setting '%s' is a path and paths cannot be empty", s.name);
@@ -467,7 +494,9 @@ static Path parsePath(const AbstractSetting & s, const std::string & str)
 template<>
 std::filesystem::path BaseSetting<std::filesystem::path>::parse(const std::string & str) const
 {
-    return parsePath(*this, str);
+    if (str == "")
+        throw UsageError("setting '%s' is a path and paths cannot be empty", name);
+    return str;
 }
 
 template<>
@@ -477,17 +506,28 @@ std::string BaseSetting<std::filesystem::path>::to_string() const
 }
 
 template<>
-std::optional<std::filesystem::path>
-BaseSetting<std::optional<std::filesystem::path>>::parse(const std::string & str) const
+AbsolutePath BaseSetting<AbsolutePath>::parse(const std::string & str) const
+{
+    return parseAbsolutePath(*this, str);
+}
+
+template<>
+std::string BaseSetting<AbsolutePath>::to_string() const
+{
+    return value.string();
+}
+
+template<>
+std::optional<AbsolutePath> BaseSetting<std::optional<AbsolutePath>>::parse(const std::string & str) const
 {
     if (str == "")
         return std::nullopt;
     else
-        return parsePath(*this, str);
+        return parseAbsolutePath(*this, str);
 }
 
 template<>
-std::string BaseSetting<std::optional<std::filesystem::path>>::to_string() const
+std::string BaseSetting<std::optional<AbsolutePath>>::to_string() const
 {
     return value ? value->string() : "";
 }
@@ -506,47 +546,9 @@ template class BaseSetting<StringSet>;
 template class BaseSetting<StringMap>;
 template class BaseSetting<std::set<ExperimentalFeature>>;
 template class BaseSetting<std::filesystem::path>;
-template class BaseSetting<std::optional<std::filesystem::path>>;
-
-PathSetting::PathSetting(
-    Config * options,
-    const Path & def,
-    const std::string & name,
-    const std::string & description,
-    const StringSet & aliases)
-    : BaseSetting<Path>(def, true, name, description, aliases)
-{
-    options->addSetting(this);
-}
-
-Path PathSetting::parse(const std::string & str) const
-{
-    return parsePath(*this, str);
-}
-
-OptionalPathSetting::OptionalPathSetting(
-    Config * options,
-    const std::optional<Path> & def,
-    const std::string & name,
-    const std::string & description,
-    const StringSet & aliases)
-    : BaseSetting<std::optional<Path>>(def, true, name, description, aliases)
-{
-    options->addSetting(this);
-}
-
-std::optional<Path> OptionalPathSetting::parse(const std::string & str) const
-{
-    if (str == "")
-        return std::nullopt;
-    else
-        return parsePath(*this, str);
-}
-
-void OptionalPathSetting::operator=(const std::optional<Path> & v)
-{
-    this->assign(v);
-}
+template class BaseSetting<AbsolutePath>;
+template class BaseSetting<std::optional<AbsolutePath>>;
+template class BaseSetting<std::optional<std::string>>;
 
 bool ExperimentalFeatureSettings::isEnabled(const ExperimentalFeature & feature) const
 {
