@@ -1,8 +1,6 @@
 #include "nix/util/util.hh"
 #include "nix/util/environment-variables.hh"
 
-extern char ** environ __attribute__((weak));
-
 namespace nix {
 
 std::optional<std::string> getEnv(const std::string & key)
@@ -21,24 +19,18 @@ std::optional<std::string> getEnvNonEmpty(const std::string & key)
     return value;
 }
 
-StringMap getEnv()
+std::optional<OsString> getEnvOsNonEmpty(const OsString & key)
 {
-    StringMap env;
-    for (size_t i = 0; environ[i]; ++i) {
-        auto s = environ[i];
-        auto eq = strchr(s, '=');
-        if (!eq)
-            // invalid env, just keep going
-            continue;
-        env.emplace(std::string(s, eq), std::string(eq + 1));
-    }
-    return env;
+    auto value = getEnvOs(key);
+    if (value == OS_STR(""))
+        return {};
+    return value;
 }
 
 void clearEnv()
 {
-    for (auto & name : getEnv())
-        unsetenv(name.first.c_str());
+    for (auto & [name, value] : getEnvOs())
+        unsetEnvOs(name.c_str());
 }
 
 void replaceEnv(const StringMap & newEnv)
