@@ -5,6 +5,7 @@
 #include "nix/store/store-api.hh"
 #include "nix/util/archive.hh"
 #include "nix/store/derivations.hh"
+#include "nix/store/globals.hh"
 #include <nlohmann/json.hpp>
 
 using namespace nix;
@@ -35,9 +36,8 @@ struct CmdAddDerivation : MixDryRun, StoreCommand
 
         auto drv = Derivation::parseJsonAndValidate(*store, json);
 
-        auto drvPath = writeDerivation(*store, drv, NoRepair, /* read only */ dryRun);
-
-        writeDerivation(*store, drv, NoRepair, dryRun);
+        auto drvPath =
+            (dryRun || settings.readOnlyMode) ? computeStorePath(*store, drv) : store->writeDerivation(drv, NoRepair);
 
         logger->cout("%s", store->printStorePath(drvPath));
     }
