@@ -145,7 +145,7 @@ struct CmdFormatterBuild : MixFormatter, MixOutLinkByDefault
         auto buildables = unresolvedApp.build(evalStore, store);
         createOutLinksMaybe(buildables, store);
 
-        logger->cout("%s", app.program);
+        logger->cout("%s", app.program.string());
     };
 };
 
