@@ -28,6 +28,7 @@
 #include "nix/util/users.hh"
 #include "nix/cmd/network-proxy.hh"
 #include "nix/cmd/compatibility-settings.hh"
+#include "nix/util/fun.hh"
 #include "man-pages.hh"
 
 using namespace nix;
@@ -140,7 +141,7 @@ static void main_nix_build(int argc, char ** argv)
     auto myName = isNixShell ? "nix-shell" : "nix-build";
 
     auto inShebang = false;
-    std::string script;
+    std::filesystem::path script;
     std::vector<std::string> savedArgs;
 
     AutoDelete tmpDir(createTempDir("", myName));
@@ -200,9 +201,9 @@ static void main_nix_build(int argc, char ** argv)
     {
         using LegacyArgs::LegacyArgs;
 
-        void setBaseDir(Path baseDir)
+        void setBaseDir(std::filesystem::path baseDir)
         {
-            commandBaseDir = baseDir;
+            commandBaseDir = baseDir.string();
         }
     };
 
@@ -284,11 +285,15 @@ static void main_nix_build(int argc, char ** argv)
                     fmt("exec %1% %2% -e 'load(ARGV.shift)' -- %3% %4%",
                         execArgs,
                         interpreter,
-                        escapeShellArgAlways(script),
+                        escapeShellArgAlways(script.string()),
                         joined.view());
             } else {
                 envCommand =
-                    fmt("exec %1% %2% %3% %4%", execArgs, interpreter, escapeShellArgAlways(script), joined.view());
+                    fmt("exec %1% %2% %3% %4%",
+                        execArgs,
+                        interpreter,
+                        escapeShellArgAlways(script.string()),
+                        joined.view());
             }
         }
 
@@ -313,15 +318,15 @@ static void main_nix_build(int argc, char ** argv)
         throw UsageError("'-p' and '-E' are mutually exclusive");
 
     auto store = openStore();
-    auto evalStore = myArgs.evalStoreUrl ? openStore(*myArgs.evalStoreUrl) : store;
+    auto evalStore = myArgs.evalStoreUrl ? openStore(StoreReference{*myArgs.evalStoreUrl}) : store;
 
-    auto state = std::make_unique<EvalState>(myArgs.lookupPath, evalStore, fetchSettings, evalSettings, store);
+    auto state = std::make_shared<EvalState>(myArgs.lookupPath, evalStore, fetchSettings, evalSettings, store);
     state->repair = myArgs.repair;
     if (myArgs.repair)
         buildMode = bmRepair;
 
     if (inShebang && compatibilitySettings.nixShellShebangArgumentsRelativeToScript) {
-        myArgs.setBaseDir(absPath(dirOf(script)));
+        myArgs.setBaseDir(absPath(script.parent_path()));
     }
     auto autoArgs = myArgs.getAutoArgs(*state);
 
@@ -373,17 +378,17 @@ static void main_nix_build(int argc, char ** argv)
         exprs = {state->parseStdin()};
     else
         for (auto i : remainingArgs) {
+            auto shebangBaseDir = absPath(script.parent_path());
             if (fromArgs) {
-                auto shebangBaseDir = absPath(dirOf(script));
                 exprs.push_back(state->parseExprFromString(
                     std::move(i),
                     (inShebang && compatibilitySettings.nixShellShebangArgumentsRelativeToScript)
-                        ? lookupFileArg(*state, shebangBaseDir)
+                        ? lookupFileArg(*state, shebangBaseDir.string())
                         : state->rootPath(".")));
             } else {
                 auto absolute = i;
                 try {
-                    absolute = canonPath(absPath(i), true);
+                    absolute = canonPath(absPath(std::filesystem::path{i}), true).string();
                 } catch (Error & e) {
                 };
                 auto [path, outputNames] = parsePathWithOutputs(absolute);
@@ -392,9 +397,10 @@ static void main_nix_build(int argc, char ** argv)
                 else {
                     /* If we're in a #! script, interpret filenames
                        relative to the script. */
-                    auto baseDir = inShebang && !packages ? absPath(i, absPath(dirOf(script))) : i;
+                    std::filesystem::path iPath{i};
+                    auto baseDir = inShebang && !packages ? absPath(iPath, &shebangBaseDir) : iPath;
 
-                    auto sourcePath = lookupFileArg(*state, baseDir);
+                    auto sourcePath = lookupFileArg(*state, baseDir.string());
                     auto resolvedPath = isNixShell ? resolveShellExprPath(sourcePath) : resolveExprPath(sourcePath);
 
                     exprs.push_back(state->parseExprFromFile(resolvedPath));
@@ -531,7 +537,7 @@ static void main_nix_build(int argc, char ** argv)
             shell = store->printStorePath(shellDrvOutputs.at("out").value()) + "/bin/bash";
         }
 
-        if (experimentalFeatureSettings.isEnabled(Xp::CaDerivations)) {
+        if (drv.shouldResolve()) {
             auto resolvedDrv = drv.tryResolve(*store);
             assert(resolvedDrv && "Successfully resolved the derivation");
             drv = *resolvedDrv;
@@ -552,7 +558,10 @@ static void main_nix_build(int argc, char ** argv)
 
         env["NIX_BUILD_TOP"] = env["TMPDIR"] = env["TEMPDIR"] = env["TMP"] = env["TEMP"] = tmpDir.path().string();
         env["NIX_STORE"] = store->storeDir;
-        env["NIX_BUILD_CORES"] = fmt("%d", settings.buildCores ? settings.buildCores : settings.getDefaultCores());
+        env["NIX_BUILD_CORES"] =
+            fmt("%d",
+                settings.getLocalSettings().buildCores ? settings.getLocalSettings().buildCores
+                                                       : settings.getDefaultCores());
 
         DerivationOptions<StorePath> drvOptions;
         try {
@@ -567,7 +576,7 @@ static void main_nix_build(int argc, char ** argv)
         for (auto & var : drv.env)
             if (drvOptions.passAsFile.count(var.first)) {
                 auto fn = ".attr-" + std::to_string(fileNr++);
-                Path p = (tmpDir.path() / fn).string();
+                auto p = (tmpDir.path() / fn).string();
                 writeFile(p, var.second);
                 env[var.first + "Path"] = p;
             } else
@@ -578,18 +587,16 @@ static void main_nix_build(int argc, char ** argv)
         if (drv.structuredAttrs) {
             StorePathSet inputs;
 
-            std::function<void(const StorePath &, const DerivedPathMap<StringSet>::ChildNode &)> accumInputClosure;
-
-            accumInputClosure = [&](const StorePath & inputDrv,
-                                    const DerivedPathMap<StringSet>::ChildNode & inputNode) {
-                auto outputs = store->queryPartialDerivationOutputMap(inputDrv, &*evalStore);
-                for (auto & i : inputNode.value) {
-                    auto o = outputs.at(i);
-                    store->computeFSClosure(*o, inputs);
-                }
-                for (const auto & [outputName, childNode] : inputNode.childMap)
-                    accumInputClosure(*outputs.at(outputName), childNode);
-            };
+            fun<void(const StorePath &, const DerivedPathMap<StringSet>::ChildNode &)> accumInputClosure =
+                [&](const StorePath & inputDrv, const DerivedPathMap<StringSet>::ChildNode & inputNode) {
+                    auto outputs = store->queryPartialDerivationOutputMap(inputDrv, &*evalStore);
+                    for (auto & i : inputNode.value) {
+                        auto o = outputs.at(i);
+                        store->computeFSClosure(*o, inputs);
+                    }
+                    for (const auto & [outputName, childNode] : inputNode.childMap)
+                        accumInputClosure(*outputs.at(outputName), childNode);
+                };
 
             for (const auto & [inputDrv, inputNode] : drv.inputDrvs.map)
                 accumInputClosure(inputDrv, inputNode);
@@ -613,6 +620,8 @@ static void main_nix_build(int argc, char ** argv)
            environment variables and shell functions.  Also don't
            lose the current $PATH directories. */
         auto rcfile = (tmpDir.path() / "rc").string();
+        auto tz = getEnv("TZ");
+        auto tzExport = tz ? "export TZ=" + escapeShellArgAlways(*tz) + "; " : "";
         std::string rc = fmt(
                 (R"(_nix_shell_clean_tmpdir() { command rm -rf %1%; };)"s
                   "trap _nix_shell_clean_tmpdir EXIT; "
@@ -644,9 +653,9 @@ static void main_nix_build(int argc, char ** argv)
                 escapeShellArgAlways(tmpDir.path().string()),
                 (pure ? "" : "p=$PATH; "),
                 (pure ? "" : "PATH=$PATH:$p; unset p; "),
-                escapeShellArgAlways(dirOf(*shell)),
+                escapeShellArgAlways(std::filesystem::path(*shell).parent_path().string()),
                 escapeShellArgAlways(*shell),
-                (getenv("TZ") ? (std::string("export TZ=") + escapeShellArgAlways(getenv("TZ")) + "; ") : ""),
+                tzExport,
                 envCommand);
         vomit("Sourcing nix-shell with file %s and contents:\n%s", rcfile, rc);
         writeFile(rcfile, rc);
