@@ -5,6 +5,7 @@
 #include "nix/expr/eval.hh"
 #include "nix/expr/get-drvs.hh"
 #include "nix/store/globals.hh"
+#include "nix/util/config-global.hh"
 #include "nix/store/names.hh"
 #include "nix/store/profiles.hh"
 #include "nix/store/path-with-outputs.hh"
@@ -34,13 +35,46 @@
 using namespace nix;
 using std::cout;
 
+/**
+ * Settings related to Nix user environments.
+ */
+struct EnvSettings : Config
+{
+    Setting<bool> keepDerivations{
+        this,
+        false,
+        "keep-env-derivations",
+        R"(
+          If `false` (default), derivations are not stored in Nix user
+          environments. That is, the derivations of any build-time-only
+          dependencies may be garbage-collected.
+
+          If `true`, when you add a Nix derivation to a user environment, the
+          path of the derivation is stored in the user environment. Thus, the
+          derivation isn't garbage-collected until the user environment
+          generation is deleted (`nix-env --delete-generations`). To prevent
+          build-time-only dependencies from being collected, you should also
+          turn on `keep-outputs`.
+
+          The difference between this option and `keep-derivations` is that
+          this one is "sticky": it applies to any user environment created
+          while this option was enabled, while `keep-derivations` only applies
+          at the moment the garbage collector is run.
+        )",
+        {"env-keep-derivations"}};
+};
+
+EnvSettings envSettings;
+
+static GlobalConfig::Register rSettings(&envSettings);
+
 typedef enum { srcNixExprDrvs, srcNixExprs, srcStorePaths, srcProfile, srcAttrPath, srcUnknown } InstallSourceType;
 
 struct InstallSourceInfo
 {
     InstallSourceType type;
     std::shared_ptr<SourcePath> nixExprPath; /* for srcNixExprDrvs, srcNixExprs */
-    Path profile;                            /* for srcProfile */
+    std::filesystem::path profile;           /* for srcProfile */
     std::string systemFilter;                /* for srcNixExprDrvs */
     Bindings * autoArgs;
 };
@@ -48,7 +82,7 @@ struct InstallSourceInfo
 struct Globals
 {
     InstallSourceInfo instSource;
-    Path profile;
+    std::filesystem::path profile;
     std::shared_ptr<EvalState> state;
     bool dryRun;
     bool preserveInstalled;
@@ -488,8 +522,8 @@ static void setMetaFlag(EvalState & state, PackageInfo & drv, const std::string
     drv.setMeta(name, v);
 }
 
-static void
-installDerivations(Globals & globals, const Strings & args, const Path & profile, std::optional<int> priority)
+static void installDerivations(
+    Globals & globals, const Strings & args, const std::filesystem::path & profile, std::optional<int> priority)
 {
     debug("installing derivations");
 
@@ -546,7 +580,7 @@ installDerivations(Globals & globals, const Strings & args, const Path & profile
         if (globals.dryRun)
             return;
 
-        if (createUserEnv(*globals.state, allElems, profile, settings.envKeepDerivations, lockToken))
+        if (createUserEnv(*globals.state, allElems, profile, envSettings.keepDerivations, lockToken))
             break;
     }
 }
@@ -657,7 +691,7 @@ static void upgradeDerivations(Globals & globals, const Strings & args, UpgradeT
         if (globals.dryRun)
             return;
 
-        if (createUserEnv(*globals.state, newElems, globals.profile, settings.envKeepDerivations, lockToken))
+        if (createUserEnv(*globals.state, newElems, globals.profile, envSettings.keepDerivations, lockToken))
             break;
     }
 }
@@ -716,7 +750,7 @@ static void opSetFlag(Globals & globals, Strings opFlags, Strings opArgs)
         checkSelectorUse(selectors);
 
         /* Write the new user environment. */
-        if (createUserEnv(*globals.state, installedElems, globals.profile, settings.envKeepDerivations, lockToken))
+        if (createUserEnv(*globals.state, installedElems, globals.profile, envSettings.keepDerivations, lockToken))
             break;
     }
 }
@@ -766,7 +800,7 @@ static void opSet(Globals & globals, Strings opFlags, Strings opArgs)
     switchLink(globals.profile, generation);
 }
 
-static void uninstallDerivations(Globals & globals, Strings & selectors, Path & profile)
+static void uninstallDerivations(Globals & globals, Strings & selectors, const std::filesystem::path & profile)
 {
     while (true) {
         auto lockToken = optimisticLockProfile(profile);
@@ -799,7 +833,7 @@ static void uninstallDerivations(Globals & globals, Strings & selectors, Path &
         if (globals.dryRun)
             return;
 
-        if (createUserEnv(*globals.state, workingElems, profile, settings.envKeepDerivations, lockToken))
+        if (createUserEnv(*globals.state, workingElems, profile, envSettings.keepDerivations, lockToken))
             break;
     }
 }
@@ -1260,7 +1294,7 @@ static void opSwitchProfile(Globals & globals, Strings opFlags, Strings opArgs)
     if (opArgs.size() != 1)
         throw UsageError("exactly one argument expected");
 
-    Path profile = absPath(opArgs.front());
+    auto profile = absPath(std::filesystem::path{opArgs.front()});
     auto profileLink = settings.useXDGBaseDirectories ? createNixStateDir() / "profile" : getHome() / ".nix-profile";
 
     switchLink(profileLink, profile);
@@ -1380,11 +1414,11 @@ static int main_nix_env(int argc, char ** argv)
 
         if (!pathExists(nixExprPath)) {
             try {
+                auto profilesDirOpts = settings.getProfileDirsOptions();
                 createDirs(nixExprPath);
-                replaceSymlink(defaultChannelsDir(), nixExprPath / "channels");
+                replaceSymlink(defaultChannelsDir(profilesDirOpts), nixExprPath / "channels");
                 if (!isRootUser())
-                    replaceSymlink(rootChannelsDir(), nixExprPath / "channels_root");
-            } catch (std::filesystem::filesystem_error &) {
+                    replaceSymlink(rootChannelsDir(profilesDirOpts), nixExprPath / "channels_root");
             } catch (Error &) {
             }
         }
@@ -1490,7 +1524,7 @@ static int main_nix_env(int argc, char ** argv)
             globals.profile = getEnv("NIX_PROFILE").value_or("");
 
         if (globals.profile == "")
-            globals.profile = getDefaultProfile().string();
+            globals.profile = getDefaultProfile(settings.getProfileDirsOptions()).string();
 
         op(globals, std::move(opFlags), std::move(opArgs));
 
