@@ -4,6 +4,8 @@
 #include "nix/store/store-api.hh"
 #include "nix/store/store-cast.hh"
 #include "nix/store/gc-store.hh"
+#include "nix/util/error.hh"
+#include "nix/util/logging.hh"
 
 using namespace nix;
 
@@ -15,7 +17,7 @@ struct CmdStoreGC : StoreCommand, MixDryRun
     {
         addFlag({
             .longName = "max",
-            .description = "Stop after freeing *n* bytes of disk space.",
+            .description = "Stop after freeing *n* bytes of disk space. Cannot be combined with --dry-run.",
             .labels = {"n"},
             .handler = {&options.maxFreed},
         });
@@ -35,11 +37,14 @@ struct CmdStoreGC : StoreCommand, MixDryRun
 
     void run(ref<Store> store) override
     {
+        if (options.maxFreed != std::numeric_limits<uint64_t>::max() && dryRun)
+            throw UsageError("options --max and --dry-run cannot be combined");
+
         auto & gcStore = require<GcStore>(*store);
 
         options.action = dryRun ? GCOptions::gcReturnDead : GCOptions::gcDeleteDead;
         GCResults results;
-        PrintFreed freed(options.action == GCOptions::gcDeleteDead, results);
+        Finally printer([&] { printFreed(dryRun, results); });
         gcStore.collectGarbage(options, results);
     }
 };
