@@ -3,6 +3,7 @@
 #include "nix/store/path-references.hh"
 #include "nix/util/source-accessor.hh"
 #include "nix/main/shared.hh"
+#include "nix/util/fun.hh"
 
 #include <queue>
 
@@ -165,12 +166,12 @@ struct CmdWhyDepends : SourceExprCommand, MixOperateOnOptions
            closure (i.e., that have a non-infinite distance to
            'dependency'). Print every edge on a path between `package`
            and `dependency`. */
-        std::function<void(Node &, const std::string &, const std::string &)> printNode;
-
         struct BailOut
         {};
 
-        printNode = [&](Node & node, const std::string & firstPad, const std::string & tailPad) {
+        fun<void(Node &, const std::string &, const std::string &)> printNode = [&](Node & node,
+                                                                                    const std::string & firstPad,
+                                                                                    const std::string & tailPad) {
             assert(node.dist != inf);
             if (precise) {
                 logger->cout(
